using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

/// <summary>
/// hide character version 4.0
/// a modification of Seamonkey's "hide character 1.0", which is a modification of Acidbubbles' "improved pov"
/// tl;dr: it adds a translucency and render queue slider for each part of the person's skin.
/// you can modify what parts are hidden by removing things under "MaterialsToHide" (the list starting with defaultMat)
/// but that's not really a problem anymore given that each part has a slider now lol
/// this edit was made by Doofenschmalphys
/// </summary>
public class HideCharacter : MVRScript
{
    private Atom _person;
    private DAZCharacterSelector _selector;
    private JSONStorableBool _hideCharacterToggleJSON;
    private JSONStorableBool _hideCharacterJSON;

    private SkinHandler _skinHandler;
    // For change detection purposes
    private DAZCharacter _character;

    private bool _lastActive;
    // Requires re-generating all shaders and materials, either because last frame was not ready or because something changed
    private bool _dirty;
    // To avoid spamming errors when something failed
    private bool _failedOnce;
    // When waiting for a model to load, how long before we abandon
    private int _tryAgainAttempts;
	private	JSONStorableFloat transp;
	private UIDynamicButton changeAll;




    public override void Init()
    {
		
        try
        {
            if (containingAtom?.type != "Person")
            {
                SuperController.LogError($"This plugin only works on Person atoms");
                DestroyImmediate(this);
                return;
            }

            _person = containingAtom;
            _selector = _person.GetComponentInChildren<DAZCharacterSelector>();

            InitControls();
            Camera.onPreRender += OnPreRender;
            Camera.onPostRender += OnPreRender;
			transp = new JSONStorableFloat("Transparency",-1f,1f,.0f,true);
			
			
			
			CreateSlider(transp,false);
			UIDynamic spacer= CreateSpacer(true);
			spacer.height=55f;
				
			changeAll = CreateButton("Copy to All Values",false);
			changeAll.button.onClick.AddListener(delegate ()
				{
						foreach(var key in Globals.globalTrans.Keys.ToList() )
						{
							Globals.globalTrans[key].val = transp.val;
						}
				});
				
			
				
				
			
			
        }
        catch (Exception e)
        {
            SuperController.LogError("something failed: " + e);
            DestroyImmediate(this);
        }
    }

    private void OnPreRender(Camera cam)
    {

        try
        {
            if (_skinHandler != null)
                _skinHandler.BeforeRender();
        }
        catch (Exception e)
        {
            if (_failedOnce) return;
            _failedOnce = true;
            SuperController.LogError("something failed on prerender: " + e);
        }
    }

    private void OnPostRender(Camera cam)
    {

        try
        {
            if (_skinHandler != null)
                _skinHandler.AfterRender();
        }
        catch (Exception e)
        {
            if (_failedOnce) return;
            _failedOnce = true;
            SuperController.LogError("something failed on postrender: " + e);
        }
    }

    private void InitControls()
    {
        try
        {
            {
                var possessedOnlyDefaultValue = true;

                _hideCharacterToggleJSON = new JSONStorableBool("Toggle off to hide character", possessedOnlyDefaultValue);
                RegisterBool(_hideCharacterToggleJSON);
                var possessedOnlyCheckbox = CreateToggle(_hideCharacterToggleJSON, true);
                possessedOnlyCheckbox.toggle.onValueChanged.AddListener(delegate (bool val)
                {
                    _dirty = true;
                });
            }

            {
                _hideCharacterJSON = new JSONStorableBool("Keep this toggled on", true);
                RegisterBool(_hideCharacterJSON);
                var hideFaceToggle = CreateToggle(_hideCharacterJSON, true);
                hideFaceToggle.toggle.onValueChanged.AddListener(delegate (bool val)
                {
                    _dirty = true;
                });
            }
        }
        catch (Exception e)
        {
            SuperController.LogError("Failed to register controls: " + e);
        }
    }

    public void OnDisable()
    {
        try
        {
            _dirty = false;
            ApplyAll(false);
            _lastActive = false;
        }
        catch (Exception e)
        {
            SuperController.LogError("something failed: " + e);
        }
    }

    public void OnDestroy()
    {
        OnDisable();
        Camera.onPreRender -= OnPreRender;
        Camera.onPostRender -= OnPostRender;
    }

    public void Update()
    {
        try
        {
            var active = !_hideCharacterToggleJSON.val;
			Globals.counter = transp;
			
			
			if(Globals.newbie != null){
				Globals.globalTrans.Add(Globals.newbie.material.name,new JSONStorableFloat(Globals.newbie.material.name,-1f,-1f,.0f,true));
			    Globals.globalQueue.Add(Globals.newbie.material.name,new JSONStorableFloat(Globals.newbie.material.name + "Render Queue",(float)Globals.newbie.material.renderQueue,-1f,5000f,true));
				CreateSlider(Globals.globalTrans[Globals.newbie.material.name],true);
				CreateSlider(Globals.globalQueue[Globals.newbie.material.name],false);
				//SuperController.LogError(Globals.newbie.material.shader.ToString());
				
				Globals.newbie = null;
			}
			
			

            if (!_lastActive && active)
            {
                ApplyAll(true);
                _lastActive = true;
            }
            else if (_lastActive && !active)
            {
                ApplyAll(false);
                _lastActive = false;
            }
            else if (_dirty)
            {
                _dirty = false;
                ApplyAll(_lastActive);
            }
            else if (_lastActive && _selector.selectedCharacter != _character)
            {
                _skinHandler?.Restore();
                _skinHandler = null;
                ApplyAll(true);
            }
        }
        catch (Exception e)
        {
            if (_failedOnce) return;
            _failedOnce = true;
            SuperController.LogError("something failed: " + e);
        }
    }

    bool once;
    private void ApplyAll(bool active)
    {
        // Try again next frame
        if (_selector.selectedCharacter?.skin == null)
        {
            MakeDirty("Skin not yet loaded.");
            return;
        }

        _character = _selector.selectedCharacter;

        if (UpdateHandler(ref _skinHandler, active && _hideCharacterJSON.val))
            ConfigureHandler("Skin", ref _skinHandler, _skinHandler.Configure(_character.skin));
        if (!_dirty) _tryAgainAttempts = 0;
    }

    private void MakeDirty(string reason)
    {
        _dirty = true;
        _tryAgainAttempts++;
        if (_tryAgainAttempts > 90 * 20) // Approximately 20 to 40 seconds
        {
            SuperController.LogError("something failed. Reason: " + reason + ".");
            enabled = false;
        }
    }

    private void ConfigureHandler<T>(string what, ref T handler, int result)
     where T : IHandler, new()
    {
        switch (result)
        {
            case HandlerConfigurationResult.Success:
                break;
            case HandlerConfigurationResult.CannotApply:
                handler = default(T);
                break;
            case HandlerConfigurationResult.TryAgainLater:
                handler = default(T);
                MakeDirty(what + " is still waiting for assets to be ready.");
                break;
        }
    }

    private bool UpdateHandler<T>(ref T handler, bool active)
     where T : IHandler, new()
    {
        if (handler == null && active)
        {
            handler = new T();
            return true;
        }

        if (handler != null && active)
        {
            handler.Restore();
            handler = new T();
            return true;
        }

        if (handler != null && !active)
        {
            handler.Restore();
            handler = default(T);
        }

        return false;
    }

    public static class HandlerConfigurationResult
    {
        public const int Success = 0;
        public const int CannotApply = 1;
        public const int TryAgainLater = 2;
    }

    public interface IHandler
    {
        void Restore();
        void BeforeRender();
        void AfterRender();
    }

    public class SkinHandler : IHandler
    {
        public class SkinShaderMaterialReference
        {
            public Material material;
            public Shader originalShader;
            public float originalAlphaAdjust;
            public float originalColorAlpha;
            public Color originalSpecColor;
			

            public static SkinShaderMaterialReference FromMaterial(Material material)
            {
                var materialRef = new SkinShaderMaterialReference();
                materialRef.material = material;
                materialRef.originalShader = material.shader;
                materialRef.originalAlphaAdjust = material.GetFloat("_AlphaAdjust");
                materialRef.originalColorAlpha = material.GetColor("_Color").a;
                materialRef.originalSpecColor = material.GetColor("_SpecColor");
		
                return materialRef;
            }
        }

        public static readonly string[] MaterialsToHide = new[]
        {
			"defaultMat",
			"Cornea",
			"Ears",
			"Eyelashes",
			"EyeReflection",
			"Face",
			"Feet",
			"Fingernails",
			"Forearms",
			"Gums",
			"Hands",
			"Head",
			"Hips",
			"InnerMouth",
			"Irises",
			"Lacrimals",
			"Legs",
			"Lips",
			"Neck",
			"Nipples",
			"Nostrils",
			"Pupils",
			"Sclera",
			"Shoulders",
			"Tear",
			"Teeth",
			"Toenails",
			"Tongue",
			"Torso"
        };

        public static IList<Material> GetMaterialsToHide(DAZSkinV2 skin)
        {

            var materials = new List<Material>(MaterialsToHide.Length);
			

            foreach (var material in skin.GPUmaterials)
            {
                if (!MaterialsToHide.Any(materialToHide => material.name.StartsWith(materialToHide)))
                    continue;

                materials.Add(material);
            }

            return materials;
        }

        private static Dictionary<string, Shader> ReplacementShaders = new Dictionary<string, Shader>
            {
                // Opaque materials
                { "Custom/Subsurface/GlossCullComputeBuff", Shader.Find("Custom/Subsurface/TransparentGlossSeparateAlphaComputeBuff") },
                { "Custom/Subsurface/GlossNMCullComputeBuff", Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff") },
                { "Custom/Subsurface/GlossNMDetailCullComputeBuff", Shader.Find("Custom/Subsurface/TransparentGlossNMDetailNoCullSeparateAlphaComputeBuff") },
                { "Custom/Subsurface/CullComputeBuff", Shader.Find("Custom/Subsurface/TransparentSeparateAlphaComputeBuff") },
				{ "Custom/Subsurface/GlossNMTessMappedFixedComputeBuff", Shader.Find("Custom/Subsurface/TransparentGlossNMSeparateAlphaComputeBuff") },
                // Transparent materials
                { "Custom/Subsurface/TransparentGlossNoCullSeparateAlphaComputeBuff", null },
                { "Custom/Subsurface/TransparentGlossComputeBuff", null },
                { "Custom/Subsurface/TransparentComputeBuff", null },
                { "Custom/Subsurface/AlphaMaskComputeBuff", null },
				//{ "Custom/Subsurface/GlossNMTessMappedFixedComputeBuff", null },
                { "Marmoset/Transparent/Simple Glass/Specular IBLComputeBuff", null },
				
            };

        private DAZSkinV2 _skin;
        private List<SkinShaderMaterialReference> _materialRefs;

        public int Configure(DAZSkinV2 skin)
        {
            _skin = skin;
            _materialRefs = new List<SkinShaderMaterialReference>();

            foreach (var material in GetMaterialsToHide(skin))
            {
#if (IMPROVED_POV)
                if(material == null)
                    throw new InvalidOperationException("Attempts to apply the shader strategy on a destroyed material.");

                if (material.GetInt(SkinShaderMaterialReference.ImprovedPovEnabledShaderKey) == 1)
                    throw new InvalidOperationException("Attempts to apply the shader strategy on a skin that already has the plugin enabled (shader key).");
#endif

                var materialInfo = SkinShaderMaterialReference.FromMaterial(material);

                Shader shader;
                if (!ReplacementShaders.TryGetValue(material.shader.name, out shader))
                    SuperController.LogError("Missing replacement shader: '" + material.shader.name + "'");
				

                if (shader != null) material.shader = shader;

                _materialRefs.Add(materialInfo);
            }

            // This is a hack to force a refresh of the shaders cache
            skin.BroadcastMessage("OnApplicationFocus", true);
            return HandlerConfigurationResult.Success;
        }

        public void Restore()
        {
            foreach (var material in _materialRefs)
                material.material.shader = material.originalShader;

            _materialRefs = null;

            // This is a hack to force a refresh of the shaders cache
            _skin.BroadcastMessage("OnApplicationFocus", true);
        }

        public void BeforeRender()
        {
            foreach (var materialRef in _materialRefs)
            {
				float t = Globals.counter.val;
				
                var material = materialRef.material;
				if(Globals.globalTrans.ContainsKey(material.name)){
					t=Globals.globalTrans[material.name].val;
					material.renderQueue = (int)Globals.globalQueue[material.name].val;
					
				}
				else{
					
					Globals.newbie = materialRef;
					
					
				
				}
				
				
                material.SetFloat("_AlphaAdjust", t);
                var color = material.GetColor("_Color");
                material.SetColor("_Color", new Color(color.r, color.g, color.b, 0f));
                material.SetColor("_SpecColor", new Color(0f, 0f, 0f, 0f));
				
				
				
				
				
            }
        }

        public void AfterRender()
        {
            foreach (var materialRef in _materialRefs)
            {
                var material = materialRef.material;
                material.SetFloat("_AlphaAdjust", materialRef.originalAlphaAdjust);
                var color = material.GetColor("_Color");
                material.SetColor("_Color", new Color(color.r, color.g, color.b, materialRef.originalColorAlpha));
                material.SetColor("_SpecColor", materialRef.originalSpecColor);
            }
        }
    }
	
	static class Globals{
		public static JSONStorableFloat counter;
		public static SkinHandler.SkinShaderMaterialReference newbie = null;
		public static Dictionary<string, JSONStorableFloat> globalQueue = new Dictionary<string, JSONStorableFloat>{};
		public static Dictionary<string, JSONStorableFloat> globalTrans = new Dictionary<string, JSONStorableFloat>
		{
			/*
			{"Tongue-1",new JSONStorableFloat("Tongue",-1f,0f,.0f,true)},
			{"Lips",new JSONStorableFloat("Lips",-1f,0f,.0f,true)},
			{"InnerMouth-1",new JSONStorableFloat("Inner Mouth",-1f,0f,.0f,true)},
			{"Gums-1",new JSONStorableFloat("Gums",-1f,0f,.0f,true)},
			{"Teeth",new JSONStorableFloat("Teeth",-1f,0f,.0f,true)},
			{"Legs",new JSONStorableFloat("Legs",-1f,0f,.0f,true)},
		    {"Hips",new JSONStorableFloat("Hips",-1f,0f,.0f,true)},
			{"Torso",new JSONStorableFloat("Torso",-1f,0f,.0f,true)},
			{"defaultMat",new JSONStorableFloat("Genitals",-1f,0f,.0f,true)},
			{"Shoulders",new JSONStorableFloat("Shoulders",-1f,0f,.0f,true)},
			{"Feet",new JSONStorableFloat("Feet",-1f,0f,.0f,true)},
			{"Hands",new JSONStorableFloat("Hands",-1f,0f,.0f,true)},
			{"Nipples",new JSONStorableFloat("Nipples",-1f,0f,.0f,true)},
			{"Forearms",new JSONStorableFloat("Forearms",-1f,0f,.0f,true)},
			{"Face",new JSONStorableFloat("Face",-1f,0f,.0f,true)},
			{"Nostrils",new JSONStorableFloat("Nostrils",-1f,0f,.0f,true)},
			{"Tears-1",new JSONStorableFloat("Tears",-1f,0f,.0f,true)},
			{"Neck",new JSONStorableFloat("Neck",-1f,0f,.0f,true)},
			{"Eyelashes-1",new JSONStorableFloat("Eyelashes",-1f,0f,.0f,true)},
			{"EyeReflection-1",new JSONStorableFloat("Eye Reflection",-1f,0f,.0f,true)},
			{"Sclera",new JSONStorableFloat("Sclera",-1f,0f,.0f,true)},
			{"Pupils",new JSONStorableFloat("Pupils",-1f,0f,.0f,true)},
			{"Ears",new JSONStorableFloat("Ears",-1f,0f,.0f,true)},
			{"Irises",new JSONStorableFloat("Irises",-1f,0f,.0f,true)},
			{"Lacrimals",new JSONStorableFloat("Tear Ducts",-1f,0f,.0f,true)},
			{"Head",new JSONStorableFloat("Head",-1f,0f,.0f,true)},
			*/
				
				
		};
		static int _getsetcounter;
		public static int getsetcounter
		{
			set{_getsetcounter=value;}
			get{return _getsetcounter;}
		}
	}
}

